/* crash-pipe: handle core file passed from stdin
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Karol Lewandowski <k.lewandowsk@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
#define _GNU_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <syslog.h>
#include <sys/prctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>

enum {
	OPT_HELP,
	OPT_SAVE_CORE,
};

const struct option opts[] = {
	{ "help", no_argument, 0, OPT_HELP },
	{ "save-core", required_argument, 0, OPT_SAVE_CORE },
	{ 0, 0, 0, 0 }
};

static char *argv0 = "";

static void usage(void)
{
	fprintf(stderr, "usage: %s [--help] --save-core FILE_NAME\n", argv0);
}

static int copy_fd_simple(int in, int out)
{
	static char buf[4096];
	int readb, remaining;

	while ((readb = read(in, buf, sizeof(buf))) > 0) {
		int n;

		for (n = 0, remaining = readb ; remaining > 0; remaining -= n) {
			n = write(out, buf, remaining);
			if (n == -1)
				return -errno;
		}
	}

	return 1;
}

static int copy_fd_splice(int in, int out)
{
	ssize_t s;
	_Bool copied_data = 0;

	const ssize_t max_splice = SSIZE_MAX;

	do {
		s = splice(in, NULL, out, NULL, max_splice, SPLICE_F_MOVE);
		// We can try to fallback to simple copy only if first splice failed
		// (ie. we have not consumed any data yet)
		if (!copied_data && s > 0)
			copied_data = 1;

	} while (s > 0);

	if (s < 0)
		return copied_data ? -errno : 0;
	else
		return 1;
}

static int save_core(const char *core_path)
{
	int fd;
	int r = 0;

	fd = open(core_path, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	if (fd == -1) {
		syslog(LOG_ERR, "crash-pipe: Unable to save core file to %s: %m\n", core_path);
		return -errno;
	}

	r = copy_fd_splice(STDIN_FILENO, fd);
	if (r == 0) {
		lseek(fd, SEEK_SET, 0);
		r = copy_fd_simple(STDIN_FILENO, fd);
	}
	if (r < 0) {
		syslog(LOG_ERR, "crash-pipe: Error while saving core file %s: %m. Removing core.\n", core_path);
		(void)unlink(core_path); // XXX check errors here too
	}

	close(fd);
	return r;
}

int main(int argc, char *argv[])
{
	int c;
	char *opt_save_core = NULL;
	int ret = 1;

	prctl(PR_SET_DUMPABLE, 0);

	argv0 = argv[0];

	while ((c = getopt_long_only(argc, argv, "", opts, NULL)) != -1) {

		if (c == OPT_HELP) {
			usage();
			exit(EXIT_SUCCESS);
		} else if (c == OPT_SAVE_CORE) {
			opt_save_core = strdup(optarg);
			if (!opt_save_core) {
				syslog(LOG_CRIT, "Out of memory. Exiting.");
				exit(EXIT_FAILURE);
			}
		}
	}

	argc -= optind;
	argv += optind;

	if (opt_save_core)
		ret = save_core(opt_save_core);

	return ret >= 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
